import os
import sys
import sqlite3
import datetime

# Remixed vgstash. This time I chose an individual file on purpose:
#
# * Only need a single module to interface with a game DB. That's awesome.
# * Keeps the backend separate from the UI, as intended.
# * SQLA and pandas are both too large and unwieldy for my intent, and deps
#   will need to be kept low if I want to be cross-platform. sqlite3's part of
#   the standard library, can't beat that.
# * GUI will be implemented in tkinter, Qt, and/or Kivy.

VERSION = "0.3b8"

PROGRESS = {
    'unbeatable': 0,
    'new': 1,
    'playing': 2,
    'beaten': 3,
    'complete': 4,
}

OWNERSHIP = {
    'unowned': 0,
    'physical': 1,
    'digital': 2,
    'both': 3,
    'member': 4
}

DEFAULT_CONFIG = {
    'db_location': os.getenv('VGSTASH_DB_LOCATION', os.path.join(os.getenv('HOME', os.curdir), '.vgstash.db')),
    'progress': os.getenv('VGSTASH_DEFAULT_PROGRESS', PROGRESS['playing']),
    'ownership': os.getenv('VGSTASH_DEFAULT_OWNERSHIP', OWNERSHIP['physical'])
}

FILTERS = {
    'allgames': "SELECT * FROM games ORDER BY system, title ASC",
    'backlog':  "SELECT * FROM games WHERE ownership > 0 AND progress > 0 AND progress < 3 ORDER BY system, title ASC",
    'borrowing': "SELECT * FROM games WHERE ownership = 0 AND progress = 2 ORDER BY system, title ASC",
    'complete': "SELECT * FROM games WHERE progress = 4 ORDER BY system, title ASC",
    'digital': "SELECT * FROM games WHERE ownership = 2 ORDER BY system, title ASC",
    'done': "SELECT * FROM games WHERE progress > 2 ORDER BY system, title ASC",
    'incomplete': "SELECT * FROM games WHERE progress = 3 AND ownership > 0 ORDER BY system, title ASC",
    'members': "SELECT * FROM games WHERE ownership = 4 ORDER BY system, title ASC",
    'new': "SELECT * FROM games WHERE progress = 1 ORDER BY system, title ASC",
    'notes': "SELECT * FROM games WHERE notes NOT LIKE '' ORDER BY system, title ASC",
    'owned': "SELECT * FROM games WHERE ownership > 0 AND ownership < 4 ORDER BY system, title ASC",
    'physical': "SELECT * FROM games WHERE ownership = 1 ORDER BY system, title ASC",
    'playlog': "SELECT * FROM games WHERE ownership > 0 AND progress = 2 ORDER BY system, title ASC",
    'unowned': "SELECT * FROM games WHERE ownership = 0 ORDER BY system, title ASC",
    'backlog_age': "SELECT title, system, ownership, ((unixepoch() - p_date) / (60 * 60 * 24 * 365)) AS years, ((unixepoch() - p_date) / (60 * 60 * 24) % 365) AS days FROM games WHERE (progress > 0 AND progress < 3) AND p_date ORDER BY years DESC, days DESC, system ASC, title ASC",
}

def iso_to_unix(t):
    return datetime.date.fromisoformat(t).strftime("%s")

def unix_to_iso(t):
    return datetime.date.fromtimestamp(t).isoformat()

def kvmatch(arg, dict_map, fallback):
    """
    Match arg against keys or values in dict_map, returning fallback if no match.

    This function performs a prefix-match against the keys in dict_map. Doing
    such iteration partially defeats the purpose of using a dictionary, but
    it offers considerable ease-of-use for the CLI and acts as a validation
    function that will always return a valid value ready for committing to the
    database.

    It is generalized to support any value that needs to be mapped to an
    integer, via the dict_map.
    """
    try:
        ret = int(arg)
    except TypeError:
        ret = fallback
    except ValueError:
        found = False
        for k in dict_map.keys():
            if k.startswith(arg):
                ret = dict_map[k]
                found = True
                break
        if not found:
            ret = fallback
    finally:
        if ret not in dict_map.values():
            ret = fallback
    return ret


def vtok(arg, dict_map):
    """
    Match an integer value to a key name in the mapping dictionary. Returns a
    string (the key name) if found, False if not found.
    """
    if isinstance(arg, int):
        for k, v in dict_map.items():
            if arg == v:
                return k
    elif isinstance(arg, str):
        for k, v in dict_map.items():
            if k.startswith(arg):
                return k
        return False


class DB(object):
    """
    The central class of vgstash. It handles everything relating to storing the
    game collection.
    """
    def __init__(self, path=DEFAULT_CONFIG['db_location']):
        """
        Initiates the DB object with a 'conn' attribute, which holds the SQLite
        connection. Additionally, the connection's 'row_factory' attribute is
        set to sqlite3.Row to allow for string *and* integer indexes. This value
        may be changed by the caller after the object is returned.
        """
        try:
            self.conn = sqlite3.connect(path)
            self.conn.row_factory = sqlite3.Row
        except sqlite3.OperationalError as e:
            print("{}: {}".format(e, path))
            exit()

    def add_filter(self, filter_name, stmt):
        """
        Adds a new filter keyword to the database. Note that values are passed
        directly in, with no escaping. Use with caution.

        'filter_name' is the name of your filter, and will be the 'name' key
        when viewed with list_filters(). usable via `vgstash list [filter name]`.
        'stmt' is a plain SELECT statement representing the internal SQLite VIEW.
        """
        if filter_name.startswith("sqlite_"):
            raise ValueError("Cannot create a filter with the 'sqlite_' prefix.")

        # The call to format() is needed because sqlite doesn't allow
        # parameterized view or table names. This may affect database
        # integrity, and may disappear in a later release.
        res = self.conn.execute("CREATE VIEW\
            IF NOT EXISTS \
            '{}'\
            AS\
            {}".format(filter_name, stmt))
        FILTERS[str(filter_name)] = str(stmt)
        return self.has_filter(filter_name)

    def add_game(self, game, update=True):
        """
        Adds a Game to the database. Returns True on success and False on
        failure.
        """
        if self.has_game(game) and update:
            return self.update_game(game, game)
        else:
            c = self.conn.execute("INSERT INTO games\
                    (title, system, ownership, progress, notes, p_date, b_date, c_date)\
                    VALUES\
                    (?, ?, ?, ?, ?, ?, ?, ?)",
                    (game.title, game.system, game.ownership, game.progress, game.notes, game.p_date, game.b_date, game.c_date))
            self.conn.commit()
            return (c.rowcount > 0)

    def create_schema(self):
        """
        Initializes the database with this basic schema:

        games:
            title (TEXT)
            system (TEXT)
            ownership (INTEGER)
            progress (INTEGER)
            notes (TEXT)
            p_date (INTEGER)
            b_date (INTEGER)
            c_date (INTEGER)

        The schema is configured to use the 'title' and 'system' columns as
        primary keys, meaning it is impossible for two games with the same name
        and system to be present in the database.

        Additionally, create_schema will add the default set of filters, which
        are considered important for an ordinary vgstash DB.
        """
        # The UNIQUE clause ensures that no two games with the same title and
        # system are added to the database. An sqlite3.IntegrityError will be
        # raised when a match occurs; clients can decide how to handle it.
        #
        # The 'rowid' field is automatically generated by sqlite, and is thus
        # omitted. External modifications to the database should be done
        # through additional tables, and should reference the 'rowid' field to
        # get more exact data manipulation. Alternatively, use the *_filter
        # methods of this class to create custom reporting filters.
        try:
            self.conn.execute("""CREATE TABLE
                IF NOT EXISTS
                games (
                    title TEXT NOT NULL,
                    system TEXT NOT NULL,
                    ownership INTEGER NOT NULL DEFAULT 1,
                    progress INTEGER NOT NULL DEFAULT 1,
                    notes TEXT DEFAULT '',
                    p_date INTEGER DEFAULT NULL,
                    b_date INTEGER DEFAULT NULL,
                    c_date INTEGER DEFAULT NULL,
                    UNIQUE (title, system) ON CONFLICT ROLLBACK
                )""")
            # setup default filters while we're here
            for name, sql in sorted(FILTERS.items()):
                self.add_filter(name, sql)
            return True
        except sqlite3.OperationalError:
            print("Table already exists, skipping.")
            return False

    def delete_filter(self, filter_name):
        with self.conn:
            # try:
            self.conn.execute("DROP VIEW IF EXISTS '{}'".format(filter_name))
            return True
            # except:
            #     sys.exc_info()
            #     print("Could not remove filter '{}'.".format((filter_name))
        return False

    def delete_game(self, game):
        """
        Deletes a game from the database. Returns True on success and False
        on failure.
        """
        if self.has_game(game):
            c = self.conn.cursor()
            c.execute("DELETE FROM games\
                    WHERE title=? AND system=?",
                    (game.title, game.system))
            self.conn.commit()
            return True
        else:
            return False

    def get_game(self, title, system):
        """
        Fetches a game's information from the database. Returns a Game object.
        """
        stmt = "SELECT * FROM games WHERE title=? AND system=?"
        res = self.conn.execute(stmt, (title, system)).fetchone()
        if bool(res):
            return Game(
                    title = res['title'],
                    system = res['system'],
                    ownership = res['ownership'],
                    progress = res['progress'],
                    notes = res['notes'],
                    p_date = res['p_date'],
                    b_date = res['b_date'],
                    c_date = res['c_date'])
        else:
            raise KeyError

    def has_game(self, game, fuzzy=False):
        """
        Returns whether or not the game is in the database.

        game  - The Game object to search for.
        fuzzy - Fuzzy search, using internal 'LIKE' and replacing the game
                title's spaces with '%' characters. Defaults to False.
        """
        if fuzzy:
            game.title = "%".join(['', game.title.replace(" ", "%"), ''])
            stmt = "SELECT * FROM games WHERE title LIKE ? AND system=?"
        else:
            stmt = "SELECT * FROM games WHERE title=? AND system=?"

        res = self.conn.execute(stmt, (game.title, game.system)).fetchone()
        # res is None if there isn't a match, which evaluates to False
        return bool(res)

    def has_filter(self, filter_name):
        return filter_name in self.list_filters().keys() \
                and filter_name in FILTERS

    def list_filters(self):
        """
        Provides an iterable of filter names and their associated SELECT
        statements.
        """
        # The 'sqlite_master' table is a built-in.
        # This returns an iterable of sqlite3.Row, which can be accessed
        # similarly to a dictionary.
        res = self.conn.execute(\
                "SELECT name,sql\
                FROM sqlite_master\
                WHERE\
                type='view'\
                ORDER BY name ASC").fetchall()
        ret = {}
        for row in res:
            ret[row['name']] = row['sql']
        # Be sure to sync with internal representation
        FILTERS = ret
        return ret

    def list_games(self, filter='allgames'):
        """
        Return a list of games that meet the filter criteria.

        If the filter exists, it returns a list of sqlite3.Row objects. If the
        filter does not exist, it returns False. If no filter is specified, it
        will return a list of all games in the database.
        """
        if filter not in FILTERS.keys():
            return False
        else:
            return self.conn.execute(FILTERS[filter]).fetchall()

    def update_filter(self, filter_name, stmt):
        """
        Updates a filter's definition within the database.

        SQLite does not have a way to update VIEWs directly, so this is a
        convenience function to make updating possible.
        """
        try:
            self.delete_filter(filter_name)
            self.add_filter(filter_name, stmt)
            return True
        except:
            return False

    def update_game(self, target, source):
        """
        Look for target in the database and (if found) update it with source's
        information. Returns True on success and False on failure.

        The same Game object may be passed as both target and source to update
        a given game's values within the database.
        """
        # don't update unless it exists
        if self.has_game(target):
            c = self.conn.cursor()
            # TODO: do this better
            c.execute("""UPDATE games\
                    SET title=?, system=?, ownership=?, progress=?, notes=?,
                    p_date=?, b_date=?, c_date=?
                    WHERE title=? AND system=?""",
                    (source.title, source.system, source.ownership, source.progress, source.notes, source.p_date, source.b_date, source.c_date, target.title, target.system))
            self.conn.commit()
            return (c.rowcount > 0)
        else:
            return False


class Game(object):
    """The core data structure of vgstash."""
    def __init__(self, title, system,
                 ownership=DEFAULT_CONFIG['ownership'],
                 progress=DEFAULT_CONFIG['progress'],
                 notes="",
                 p_date="", b_date="", c_date=""):

        self.title = title
        self.system = system
        self.ownership = kvmatch(ownership, OWNERSHIP, DEFAULT_CONFIG['ownership'])
        self.progress = kvmatch(progress, PROGRESS, DEFAULT_CONFIG['progress'])
        self.notes = notes
        self.p_date = p_date
        self.b_date = b_date
        self.c_date = c_date



import vgstash
import sqlite3
import click
import os
import subprocess
import sys
import tempfile
import yaml
import json

# Click also has this, but it doesn't support a fallback value.
from shutil import get_terminal_size

def get_db():
    """Fetch a vgstash DB object from the default location.

    Change DEFAULT_CONFIG['db_location'] before calling this function
    to alter behavior."""
    return vgstash.DB(vgstash.DEFAULT_CONFIG['db_location'])


@click.group('vgstash')
def cli():
    pass


@cli.command()
def init():
    db = get_db()
    click.echo("Initializing the database...")
    if db.create_schema():
        click.echo("Schema created.")
    else:
        raise sqlite3.OperationalError("Cannot create schema.")

def row_format(row, width, header):
    """
    Prints a row from the result set into a nice table.
    """
    # The magic number comes from:
    #    3 chars per separator (9 chars total)
    #    8 chars for "System" (so there's an additional space on each side)
    #    3 chars for "Own"
    #    9 chars for "Progress" and an additional space
    #    Total is 29 characters
    twidth = int(width) - 29
    if header == True:
        click.echo("{:<{w}s} | {:^8s} | {:^3s} | {}".format(
            "Title",
            "System",
            "Own",
            "Progress",
            w=twidth)
        )
        click.echo("-" * int(width))

    titlestr = "{: <{w}s}".format(row['title'][:twidth], w=twidth)
    systemstr = "{: ^8s}".format(row['system'][:8])
    # unowned, physical, digital, both
    ownltr = [' ', 'P', '  D', 'P D', ' M']
    ownstr = "{: <3s}".format(ownltr[row['ownership']])
    progltr = {
        0: '',
        1: 'N',
        2: 'P',
        3: 'B',
        4: 'C'
    }
    progstr = "{}".format((" " * (row['progress'] - 1) * 2) + progltr[row['progress']])
    print(" | ".join((titlestr, systemstr, ownstr, progstr)))


@cli.command('add')
@click.argument('title', type=str)
@click.argument('system', type=str)
@click.argument('ownership', type=str, required=False, default=vgstash.DEFAULT_CONFIG['ownership'])
@click.argument('progress', type=str, required=False, default=vgstash.DEFAULT_CONFIG['progress'])
@click.argument('notes', type=str, required=False, default="")
@click.argument('p_date', type=str, required=False, default="")
@click.argument('b_date', type=str, required=False, default="")
@click.argument('c_date', type=str, required=False, default="")
def add(title, system, ownership, progress, notes, p_date, b_date, c_date):
    db = get_db()
    game = vgstash.Game(title, system, ownership, progress, notes, p_date, b_date, c_date)
    try:
        # Convert user-input (meant to be RFC2822/ISO8601 dates) to timestamps
        if game.p_date:
            game.p_date = vgstash.iso_to_unix(game.p_date)
        if game.b_date:
            game.b_date = vgstash.iso_to_unix(game.b_date)
        if game.c_date:
            game.c_date = vgstash.iso_to_unix(game.c_date)
        db.add_game(game, update=False)
        own_clause = (
            "do not own it",
            "physically own it",
            "digitally own it",
            "digitally and physically own it",
            "own it in a collection"
        )
        progress_clause = (
            "cannot beat",
            "haven't started",
            "are playing",
            "have beaten",
            "have completed",
        )
        note_clause = "" if len(game.notes) == 0 else " It has notes."
        date_clause = "" if not (game.p_date or game.b_date or game.c_date) else " It has date data."
        click.echo("Added {} for {}. You {} and {} it.{}{}".format(
            game.title,
            game.system,
            own_clause[game.ownership],
            progress_clause[game.progress],
            note_clause,
            date_clause
        ))
    except sqlite3.IntegrityError as e:
        print(e)
        click.echo("Couldn't add game.")


@cli.command('list')
@click.argument('filter', type=click.Choice(vgstash.FILTERS.keys()), required=False, default="allgames")
@click.option('--raw', '-r', is_flag=True, show_default=True, default=False, help="Output raw, pipe-delimited lines")
def list_games(filter, raw):
    db = get_db()
    row_data = db.list_games(filter)
    if len(row_data) == 0:
        click.echo("No games matching this filter in your collection.")
        return

    if raw:
        for r in row_data:
            l = []
            for c in r:
                if c == None:
                    l.append(str(''))
                elif type(c) == int:
                    l.append(str(c))
                else:
                    tc = c.replace('\n', '\\n')
                    tc = tc.replace('\r', '\\r')
                    l.append(tc)
            click.echo("|".join(l))
        return

    # Get column names, and a list of widths ready to go with them
    columns = row_data[0].keys()
    widths = []
    for i in range(len(columns)):
        widths.append(len(columns[i]))

    # Make a cache to manipulate the data with
    row_cache = []
    for r in row_data:
        cache_row = []
        for c in r:
            cache_row.append(c)
        row_cache.append(cache_row)
    # We should have a full, mutable cache now!


    for r in row_cache:
        for i in range(len(columns)):
            # possible values for r[i]:
            #     title -> str
            #     system -> str
            #     ownership -> int
            #     progress -> int
            #     *_date -> int
            # process fields that need massaging for display
            if r[i] == None:
                r[i] = ""
            else:
                if columns[i] == "progress":
                    if r[i] == 0:
                        r[i] = ''
                    else:
                        r[i] = vgstash.vtok(r[i], vgstash.PROGRESS)[0].capitalize()
                if columns[i] == "ownership":
                    if r[i] == 0:
                        r[i] = ''
                    else:
                        r[i] = vgstash.vtok(r[i], vgstash.OWNERSHIP)[0].capitalize()
                if columns[i] == ("p_date") and r[i] != "":
                    r[i] = vgstash.unix_to_iso(int(r[i]))
                if columns[i] == ("b_date") and r[i] != "":
                    r[i] = vgstash.unix_to_iso(int(r[i]))
                if columns[i] == ("c_date") and r[i] != "":
                    r[i] = vgstash.unix_to_iso(int(r[i]))
                if columns[i] == "notes" and len(r[i]) > 0:
                    r[i] = "*"
            # if isinstance(r[i], int):
            #     r[i] = str(r[i])
            # Store width in relevant list
            w = len(str(r[i]))
            if w > widths[i]:
                widths[i] = w

    # print the top header
    l = []
    left_fst = "{: <{w}s}"
    right_fst = "{: >{w}s}"
    center_fst = "{: ^{w}s}"

    for i in range(len(columns)):
        l.append(center_fst.format(columns[i], w=widths[i]))
    click.echo(" | ".join(l))

    l = []
    for w in widths:
        l.append("-"*w)
    click.echo("-+-".join(l))

    # print the collection now that the hard part is done!
    for r in row_cache:
        l = []
        for i in range(len(columns)):
            # TODO: set different fstring based on column name
            if columns[i] == 'years' or columns[i] == 'days':
                l.append(right_fst.format(str(r[i]), w=widths[i]))
            else:
                l.append(left_fst.format(r[i], w=widths[i]))
        click.echo(" | ".join(l))
    
@cli.command('delete')
@click.argument('title', required=True)
@click.argument('system', required=True)
def delete_game(title, system):
    db = get_db()
    target_game = vgstash.Game(title, system)
    if db.delete_game(target_game):
        click.echo("Removed {} for {} from your collection.".format(title, system))
    else:
        click.echo("That game does not exist in your collection. Please try again.")


@cli.command('update')
@click.argument('title', required=True)
@click.argument('system', required=True)
@click.argument('attr', type=click.Choice(['title', 'system', 'ownership', 'progress', 'p_date', 'b_date', 'c_date']), required=True)
@click.argument('val', required=True)
def update_game(title, system, attr, val):
    """
    Update a specific game's attribute in the database.

    Use the title and system parameters to target the game to update.
    Ownership and progress can be their letter or numeric values.
    p_date, b_date, and c_date MUST be ISO8601 dates, i.e. 2025-07-27
    """
    # TODO: Consider namedtuple as a solution
    db = get_db()
    try:
        target_game = db.get_game(title, system)
    except:
        click.echo("Game not found. Please try again.")
        return
    if attr == 'ownership':
        val = vgstash.vtok(val, vgstash.OWNERSHIP)
    if attr == 'progress':
        val = vgstash.vtok(val, vgstash.PROGRESS)
    updated_game = vgstash.Game(
        title = val if attr == 'title' else target_game.title,
        system = val if attr == 'system' else target_game.system,
        ownership = val if attr == 'ownership' else target_game.ownership,
        progress = val if attr == 'progress' else target_game.progress,
        notes = target_game.notes,
        p_date = vgstash.iso_to_unix(val) if attr == 'p_date' else target_game.p_date,
        b_date = vgstash.iso_to_unix(val) if attr == 'b_date' else target_game.b_date,
        c_date = vgstash.iso_to_unix(val) if attr == 'c_date' else target_game.c_date
    )
    if db.update_game(target_game, updated_game):
        click.echo("Updated {} for {}. Its {} is now {}.".format(title, system, attr, val))


@cli.command('notes')
@click.argument('title', required=True)
@click.argument('system', required=True)
@click.option('--edit', '-e', is_flag=True, default=False)
def notes(title, system, edit):
    db = get_db()
    try:
        target_game = db.get_game(title, system)
    except:
        click.echo("Game not found. Please try again.")
        return

    if edit:
        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(target_game.notes.encode("UTF-8"))
            tmpfile.flush()
            pre_stat = os.stat(tmpfile.name)
            o_mtime = pre_stat.st_mtime
            o_size = pre_stat.st_size
            try:
                process = subprocess.run([os.getenv("EDITOR", "vim"), tmpfile.name])
            except:
                click.echo("Could not run editor. Is it set correctly?")
                return
            tmpfile.flush()
            tmpfile.seek(0)
            post_stat = os.stat(tmpfile.name)
            n_mtime = post_stat.st_mtime
            n_size = post_stat.st_size
            note_arr = []
            for line in tmpfile:
                note_arr.append(line.decode("UTF-8").rstrip("\r\n"))
            target_game.notes = "\n".join(note_arr)
            db.update_game(target_game, target_game)
            if process.returncode == 0:
                # determine what actually happened
                if o_mtime == n_mtime:
                    if o_size == n_size:
                        click.echo("Notes for {} on {} left unchanged.".format(target_game.title, target_game.system))
                elif n_mtime > o_mtime and o_size != n_size:
                    click.echo("Notes for {} on {} have been updated!".format(target_game.title, target_game.system))
                return
            else:
                click.echo("The editor crashed. Please try again.")
                return
    else:
        if len(target_game.notes) > 0:
            click.echo("Notes for {} on {}:".format(target_game.title, target_game.system))
            click.echo()
            click.echo(target_game.notes)
        else:
            click.echo("No notes for {} on {}.".format(target_game.title, target_game.system))


@cli.command("import")
@click.option("--format", "-f", type=click.Choice(["yaml", "json"]), required=False, default="yaml")
@click.option("--update", "-u", is_flag=True, default=False, help="Overwrite existing games with the file's data")
@click.argument("filepath",
                type=click.Path(
                    readable=True,
                    resolve_path=True,
                    dir_okay=False,
                    file_okay=True),
                default=sys.stdin,
                required=False,
                )
def import_file(format, filepath, update):
    """
    Import game data from an external file matching the chosen format.

    The default format is YAML.

    Available formats:

    * JSON
    * YAML
    """
    with open(filepath) as fp:
        if format == "yaml":
            data = yaml.safe_load(fp)
        if format == "json":
            data = json.load(fp)
    db = get_db()
    count = len(data)
    for game in data:
        try:
            db.add_game(
                vgstash.Game(
                    game["title"],
                    game["system"],
                    game["ownership"],
                    game["progress"],
                    game["notes"]
                ),
                update=update
            )
        except sqlite3.IntegrityError as e:
            # skip games that already exist
            count -= 1
    if count > 0:
        click.echo("Successfully imported {} games from {}.".format(count, filepath))
    else:
        click.echo("Couldn't import any games. Is the file formatted correctly?")


@cli.command("export")
@click.option("--format", "-f", type=click.Choice(["yaml", "json"]), required=False, default="yaml")
@click.argument("filepath",
                type=click.Path(
                    exists=False,
                    readable=True,
                    writable=True,
                    resolve_path=True,
                    dir_okay=False,
                    file_okay=True),
                required=False,
                )
def export_file(format, filepath):
    """
    Export the game database to a file written in the chosen format.

    The default format is YAML.

    Available formats:

    * JSON
    * YAML
    """
    db = get_db()
    data = db.list_games()
    game_set = []
    # Time to re-read the master branch's code
    for game in data:
        g = {}
        for field in game.keys():
            g.update({field: game[field]})
        game_set.append(g)
    if not filepath:
        if format == "yaml":
            yaml.dump(game_set, sys.stdout, default_flow_style=False,
                             indent=4, allow_unicode=True)
        if format == "json":
            json.dump(game_set, sys.stdout, allow_nan=False, indent=1, skipkeys=True, sort_keys=True)
    else:
        with open(filepath, "w") as fp:
            if format == "yaml":
                yaml.dump(game_set, fp, default_flow_style=False,
                                 indent=4, allow_unicode=True)
            if format == "json":
                json.dump(game_set, fp, allow_nan=False, indent=1, skipkeys=True, sort_keys=True)
    if len(game_set) > 0:
        if filepath:
            click.echo("Successfully exported {} games to {}.".format(len(game_set), filepath))
    else:
        click.echo("Could not export any games; have you made sure your collection has games in it?")




import click
import os
import pytest
import vgstash
import vgstash_cli

from click.testing import CliRunner

verbose = True
interactive = False

# Change this to suit your testing environment
if not interactive:
    os.environ['EDITOR'] = "cat"
else:
    if not os.getenv("EDITOR"):
        os.environ['EDITOR'] = "vim"

def test_init():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['init'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Initializing the database...\nSchema created.\n"


def test_add_minimum():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['add', 'Super Mario Bros.', 'NES'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Added Super Mario Bros. for NES. You physically own it and are playing it.\n"


def test_add_ownership():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['add', 'The Legend of Zelda', 'NES', 'd'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Added The Legend of Zelda for NES. You digitally own it and are playing it.\n"


def test_add_typical():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['add', 'Sonic the Hedgehog 2', 'Genesis', '0', '3'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Added Sonic the Hedgehog 2 for Genesis. You do not own it and have beaten it.\n"


def test_add_full():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['add', 'Vectorman', 'Genesis', 'u', 'b', 'beep'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Added Vectorman for Genesis. You do not own it and have beaten it. It also has notes.\n"


def test_add_full_note_with_newline():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['add', 'Vectorman 2', 'Genesis', 'p', 'p', 'beep\nboop'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Added Vectorman 2 for Genesis. You physically own it and are playing it. It also has notes.\n"


def test_list():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.list_games, ['--raw'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == '\n'.join((
        'Sonic the Hedgehog 2|Genesis|0|3|',
        'Vectorman|Genesis|0|3|beep',
        'Vectorman 2|Genesis|1|2|beep\\nboop',
        'Super Mario Bros.|NES|1|2|',
        'The Legend of Zelda|NES|2|2|\n',
    ))


def test_list_filter():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['list', '-r', 'playlog'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == '\n'.join((
        'Vectorman 2|Genesis|1|2|beep\\nboop',
        'Super Mario Bros.|NES|1|2|',
        'The Legend of Zelda|NES|2|2|\n',
    ))


def test_list_filter_invalid():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['list', '-r', 'derp'])
    if verbose:
        print(result.output)
    assert result.exit_code != 0


def test_list_pretty():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['list', '-w', '80'])
    if verbose:
        print()
        print(result.output)
    assert result.exit_code == 0
    assert result.output == '\n'.join((
        'Title                                               |  System  | Own | Progress',
        '--------------------------------------------------------------------------------',
        'Sonic the Hedgehog 2                                | Genesis  |     |     B',
        'Vectorman                                           | Genesis  |     |     B',
        'Vectorman 2                                         | Genesis  | P   |   P',
        'Super Mario Bros.                                   |   NES    | P   |   P',
        'The Legend of Zelda                                 |   NES    |   D |   P\n',
    ))


def test_list_pretty_smaller():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['list', '-w', '60'])
    if verbose:
        print()
        print(result.output)
    assert result.exit_code == 0
    assert result.output == '\n'.join((
        'Title                           |  System  | Own | Progress',
        '------------------------------------------------------------',
        'Sonic the Hedgehog 2            | Genesis  |     |     B',
        'Vectorman                       | Genesis  |     |     B',
        'Vectorman 2                     | Genesis  | P   |   P',
        'Super Mario Bros.               |   NES    | P   |   P',
        'The Legend of Zelda             |   NES    |   D |   P\n'
    ))


def test_list_pretty_tiny():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['list', '-w', '50'])
    if verbose:
        print()
        print(result.output)
    assert result.exit_code == 0
    assert result.output == '\n'.join((
        'Title                 |  System  | Own | Progress',
        '--------------------------------------------------',
        'Sonic the Hedgehog 2  | Genesis  |     |     B',
        'Vectorman             | Genesis  |     |     B',
        'Vectorman 2           | Genesis  | P   |   P',
        'Super Mario Bros.     |   NES    | P   |   P',
        'The Legend of Zelda   |   NES    |   D |   P\n'
    ))


def test_delete():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['delete', 'Vectorman', 'Genesis'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Removed Vectorman for Genesis from your collection.\n"


def test_delete_invalid():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['delete', 'Vectorman 3', 'Genesis'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "That game does not exist in your collection. Please try again.\n"


def test_update():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['update', 'Super Mario Bros.', 'NES', 'progress', 'c'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == 'Updated Super Mario Bros. for NES. Its progress is now complete.\n'

    list_result = runner.invoke(vgstash_cli.cli, ['list', '-w', '40'])
    if verbose:
        print(list_result.output)
    assert list_result.exit_code == 0
    assert list_result.output == "\n".join((
        'Title       |  System  | Own | Progress',
        '----------------------------------------',
        'Sonic the H | Genesis  |     |     B',
        'Vectorman 2 | Genesis  | P   |   P',
        'Super Mario |   NES    | P   |       C',
        'The Legend  |   NES    |   D |   P\n'
    ))

def test_update_invalid():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['update', 'Zelda: Skyward Sword', 'NES', 'progress', 'c'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == 'Game not found. Please try again.\n'


def test_notes():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['notes', 'Vectorman 2', 'Genesis'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "\n".join((
        'Notes for Vectorman 2 on Genesis:',
        '',
        'beep',
        'boop\n'
    ))


def test_notes_invalid():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['notes', 'Zelda: Skyward Sword', 'NES', '-e'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == 'Game not found. Please try again.\n'


def test_notes_unmodified():
    runner = CliRunner()
    print("--- WHAT EDITOR SEES ---")
    result = runner.invoke(vgstash_cli.cli, ['notes', 'Vectorman 2', 'Genesis', '-e'])
    print("--- END WHAT EDITOR SEES ---")
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == 'Notes for Vectorman 2 on Genesis left unchanged.\n'


def test_notes_bad_editor():
    os.environ['EDITOR'] = "zlurp"
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['notes', 'Vectorman 2', 'Genesis', '-e'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Could not run editor. Is it set correctly?\n"


def test_notes_empty():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['notes', 'Super Mario Bros.', 'NES'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "No notes for Super Mario Bros. on NES.\n"


def test_notes_edit():
    if not interactive:
        return
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ['notes', 'Vectorman 2', 'Genesis', '-e'])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Notes for Vectorman 2 on Genesis have been updated!\n"

    # List the results to make sure they match what the editor has.
    list_runner = CliRunner()
    list_result = runner.invoke(vgstash_cli.cli, ['list', '-r'])
    if verbose:
        print(list_result.output)
    assert list_result.exit_code == 0


def test_import_file_json():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["import", "tests/data/test_import.json"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Successfully imported 2 games from {}.\n".format(os.path.join(os.getcwd(), "tests/data/test_import.json"))

    # List the results to make sure they match what the editor has.
    list_runner = CliRunner()
    list_result = runner.invoke(vgstash_cli.cli, ['list', '-w', '40'])
    if verbose:
        print(list_result.output)
    assert list_result.exit_code == 0


def test_import_file_json_update():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["import", "tests/data/test_import.json", "-u"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Successfully imported 3 games from {}.\n".format(os.path.join(os.getcwd(), "tests/data/test_import.json"))

    # List the results to make sure they match what the editor has.
    list_runner = CliRunner()
    list_result = runner.invoke(vgstash_cli.cli, ['list', '-w', '40'])
    if verbose:
        print(list_result.output)
    assert list_result.exit_code == 0


def test_import_file_yaml():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["import", "tests/data/test_import.yml"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Successfully imported 2 games from {}.\n".format(os.path.join(os.getcwd(), "tests/data/test_import.yml"))

    # List the results to make sure they match what the editor has.
    list_runner = CliRunner()
    list_result = runner.invoke(vgstash_cli.cli, ['list', '-w', '40'])
    if verbose:
        print(list_result.output)
    assert list_result.exit_code == 0


def test_import_file_yaml_update():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["import", "tests/data/test_import.yml", "-u"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Successfully imported 3 games from {}.\n".format(os.path.join(os.getcwd(), "tests/data/test_import.yml"))

    # List the results to make sure they match what the editor has.
    list_runner = CliRunner()
    list_result = runner.invoke(vgstash_cli.cli, ['list', '-w', '40'])
    if verbose:
        print(list_result.output)
    assert list_result.exit_code == 0

def test_export_no_file():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["export"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == """-   notes: ''
    ownership: 2
    progress: 4
    system: 3DS
    title: The Legend of Zelda
-   notes: ''
    ownership: 1
    progress: 3
    system: GBA
    title: Fire Emblem
-   notes: ''
    ownership: 0
    progress: 3
    system: Genesis
    title: Sonic the Hedgehog 2
-   notes: 'beep

        boop'
    ownership: 1
    progress: 2
    system: Genesis
    title: Vectorman 2
-   notes: ''
    ownership: 1
    progress: 3
    system: NES
    title: Super Mario Bros.
-   notes: ''
    ownership: 2
    progress: 2
    system: NES
    title: The Legend of Zelda
-   notes: ''
    ownership: 2
    progress: 4
    system: Switch
    title: Puyo Puyo Tetris
-   notes: ''
    ownership: 2
    progress: 3
    system: WiiU
    title: Super Smash Bros for WiiU
"""

def test_export_file_json():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["export", "-f", "json", "tests/data/test_export.json"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Successfully exported 8 games to {}.\n".format(os.path.join(os.getcwd(), "tests/data/test_export.json"))


def test_export_file_yaml():
    runner = CliRunner()
    result = runner.invoke(vgstash_cli.cli, ["export", "-f", "yaml", "tests/data/test_export.yml"])
    if verbose:
        print(result.output)
    assert result.exit_code == 0
    assert result.output == "Successfully exported 8 games to {}.\n".format(os.path.join(os.getcwd(), "tests/data/test_export.yml"))



import os
import pytest
import vgstash
import sqlite3

def test_config():
    assert vgstash.DEFAULT_CONFIG['db_location']
    assert vgstash.DEFAULT_CONFIG['progress'] in vgstash.PROGRESS.values()
    assert vgstash.DEFAULT_CONFIG['ownership'] in vgstash.OWNERSHIP.values()

@pytest.fixture(scope="module")
def vgstash_db():
    vgstash.DEFAULT_CONFIG['db_location'] = '.vgstash.db'
    yield vgstash.DB(vgstash.DEFAULT_CONFIG['db_location'])
    os.remove(vgstash.DEFAULT_CONFIG['db_location'])

def test_game_min():
    game = vgstash.Game("test_game1", "system3")
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_game_ownership():
    game = vgstash.Game("test_game2", "system3", 1)
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_game_ownership_str():
    game = vgstash.Game("test_game2", "system3", 'd')
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_game_progress():
    game = vgstash.Game("test_game3", "system3", progress=1)
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_game_progress_str():
    game = vgstash.Game("test_game3", "system3", progress='c')
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_game_notes_no_own_or_progress():
    game = vgstash.Game("test_game4", "system3", notes="Hello world")
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_game_full():
    game = vgstash.Game("test_game5", "system3", 'b', 2, "Blah")
    assert isinstance(game, vgstash.Game)
    assert isinstance(game.title, str)
    assert isinstance(game.system, str)
    assert isinstance(game.ownership, int)
    assert isinstance(game.progress, int)
    assert isinstance(game.notes, str)
    assert game.ownership in vgstash.OWNERSHIP.values()
    assert game.progress in vgstash.PROGRESS.values()

def test_db(vgstash_db):
    assert isinstance(vgstash_db.conn, sqlite3.Connection)

def test_db_create_schema(vgstash_db):
    assert vgstash_db.create_schema()

def test_db_add_game(vgstash_db):
    game = vgstash.Game("db_add_game", "system")
    assert vgstash_db.add_game(game)
    assert vgstash_db.has_game(game)

def test_db_add_game_ownership(vgstash_db):
    game = vgstash.Game("db_add_game_ownership", "system2", 'p')
    assert vgstash_db.add_game(game)
    assert vgstash_db.has_game(game)

def test_db_add_game_notes(vgstash_db):
    game = vgstash.Game("db_add_game_notes", "system2", '-', '-', 'my notes')
    assert vgstash_db.add_game(game)
    assert vgstash_db.has_game(game)

def test_db_update_game(vgstash_db):
    oldgame = vgstash.Game("db_add_game", "system")
    newgame = vgstash.Game("db_update_game", "system")
    if vgstash_db.has_game(oldgame):
        assert vgstash_db.update_game(oldgame, newgame)
        assert vgstash_db.has_game(newgame)

def test_db_delete_game(vgstash_db):
    game = vgstash.Game("db_delete_game", "system2")
    vgstash_db.add_game(game)
    assert vgstash_db.delete_game(game)

def test_db_list_games(vgstash_db):
    res = vgstash_db.list_games()
    assert isinstance(res, list)
    assert isinstance(res[0], sqlite3.Row)

def test_db_list_games_not_found(vgstash_db):
    res = vgstash_db.list_games("derp")
    assert res == False

def test_db_add_filter(vgstash_db):
    assert vgstash_db.add_filter("db_add_filter", "SELECT * FROM games WHERE system = 'system2'")
    assert vgstash_db.has_filter("db_add_filter")

def test_db_update_filter(vgstash_db):
    assert vgstash_db.update_filter("db_add_filter", "SELECT * FROM games WHERE system='system'")
    assert vgstash_db.has_filter("db_add_filter")
    assert "'system'" in vgstash.FILTERS["db_add_filter"]

def test_db_delete_filter(vgstash_db):
    assert vgstash_db.delete_filter("db_add_filter")

def test_db_list_filters(vgstash_db):
    assert len(vgstash_db.list_filters()) > 0



# -*- coding: utf-8 -*-

# Learn more: https://github.com/kennethreitz/setup.py

from setuptools import setup, find_packages


with open('README.md') as f:
    readme = f.read()

setup(
    name='vgstash',
    version='0.3b8',
    description='a video game collection management module, backed by SQLite',
    long_description=readme,
    long_description_content_type="text/markdown; variant=CommonMark",
    author='zlg',
    license='License-Expression: AGPL-3.0-only',
    author_email='zlg+vgstash@zlg.space',
    url='https://git.zlg.space/vgstash',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    py_modules=['vgstash_cli'],
    entry_points={
        'console_scripts': [
            'vgstash=vgstash_cli:cli',
            'vgstash_tk=vgstash_tk:main'
        ],
    },
    install_requires=[
        'Click>=6.0', # for CLI
        'PyYAML', # import/export YAML files
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Games/Entertainment",
        "Topic :: Utilities",
    ],
    project_urls={
        'Source': 'https://git.zlg.space/vgstash',
    }
)

